freeze;

///////////////////////////////////////////////////////////////////////////////
/////////    Copyright 2010 Jaume Pernas directed by Merce Villanueva   ///////
/////////       This program is distributed under the terms of GNU      ///////
/////////               General Public License                          ///////
///////////////////////////////////////////////////////////////////////////////
/******************************************************************************
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
******************************************************************************/
/******************************************************************************
*******************************************************************************
		CodesOverZ4 Library
		file: CodesOverZ4.m
		Library to create and manipulate Reed Muller Codes Over Z4	
		Combinatorics and Coding Group  (http://ccg.uab.es/) 
     	Autonomous University Of Barcelona (http://www.uab.cat)		
		Author: Jaume Pernas Vallès  (jpernas@deic.uab.es)
		Last Date: 27-01-2010		version 1.1
*******************************************************************************
******************************************************************************/
//TODO: Uncomment freeze when package finished
//freeze;
/*****************************************************************************
*************************   GLOBAL VARIABLES   ******************************* 
******************************************************************************/
Z  := IntegerRing();
Z2 := Integers(2);
Z4 := Integers(4);
/*****************************************************************************/
/******************************************************************************
** RemoveZeroRows function
******************************************************************************/
/**	Description:
	Given a matrix over Z4, remove the all-zero rows.
**/

RemoveZeroRows := function(G)
	ring := BaseRing(G);
	n := Ncols(G);
	m := Nrows(G);
	zero :=	ZeroMatrix(ring, 1, n);
	return Matrix([G[i]: i in [1..m]|G[i] ne zero]);
end function;

/*****************************************************************************/
/******************************************************************************
** RepetitonCodeZ4 function
******************************************************************************/
/**	Description:
	Create a code over Z4 of length n with only the all-two codeword.
**/

RepetitionCodeZ4 := function(n)
	M := RSpace(Z4,n);
	v := M![2: x in [1..n]];
	return LinearCode<Z4, n | v >;		
end function;

/*****************************************************************************/
/******************************************************************************
** RemoveOrder2Rows function
******************************************************************************/
/**	Description:
	Given a matrix over Z4, remove the rows of order two.
**/

RemoveOrder2Rows := function(G)
	error if BaseRing(G) ne Z4,
		"Error(RemoveOrder2Rows): matrix G must be over Integers(4)";
		
	list := [];
	zero := ZeroMatrix(Z4, 1, Ncols(G));
		
	for i:=1 to Nrows(G) do
		if G[i] + G[i] ne zero then 
			Append(~list, G[i]);
		end if;
	end for;
	
	if IsEmpty(list)  then
		return zero;
	else
		return Matrix(list);	
	end if;
end function;

/*****************************************************************************/
/******************************************************************************
** Replace2with1 function
******************************************************************************/
/**	Description:
	Given a matrix over Z4, replace the twos with ones in rows of order two. 
**/

Replace2with1 := function(G)
	error if BaseRing(G) ne Z4,
		"Error(Replace2with1): matrix G must be over Integers(4)";
		
	m  := Nrows(G);
	n  := Ncols(G);
	zero := ZeroMatrix(Z4, 1, n);
	
	for i:=1 to m do
		if G[i]+G[i] eq zero then 
			for j:=1 to n do
				if G[i][j] eq 2 then
					G[i][j] := 1;
				end if;
			end for;
		end if;
	end for;
	
	return G;
end function;

/*****************************************************************************/
/******************************************************************************
** MinRowsGeneratorMatrix function
******************************************************************************/

intrinsic MinRowsGeneratorMatrix(code::CodeLinRng) -> ModMatRngElt
{
	A generator matrix for the code over Z4 of type 2^gamma 4^delta, with the 
minimum number of rows, that is with gamma + delta rows: gamma rows of order 
two and delta rows of order four.
}
	Rs,f := StandardForm(code);
	Gs := GeneratorMatrix(Rs);
	m := Nrows(Gs);
	delta := 1;
	while((delta le m) and (not IsZero(2*Gs[delta]))) do
		delta := delta+1;
	end while;
	Gmin := Matrix(Z4,	[Eltseq(Rs.i@@f): i in [delta..m]] cat 
	 					[Eltseq(Rs.i@@f):i in [1..delta-1]]		);
	gamma := m - delta + 1;
	
	return Gmin,gamma,delta - 1;		
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** ReedMullerCodeQRMZ4 function
******************************************************************************/

intrinsic ReedMullerCodeQRMZ4(r::RngIntElt,m::RngIntElt) -> CodeLinRng
{
	Given an integer m >= 2 and an integer r such that 0 <= r <= m, return the 
r-th order Reed-Muller code over Z4 of length 2^m. The binary image under the 
modulo 2 map is the binary linear r-th order Reed-Muller code of length 2^m.
For r = 1 and r = m - 2, the function returns the quaternary linear Kerdock and 
Preparata code, respectively.
}
	return ReedMullerCodeZ4(r,m);		
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** PlotkinSum function
******************************************************************************/

intrinsic PlotkinSum(A::Mtrx,B::Mtrx) -> Mtrx
{
	Given matrices A and B both over the same ring and with the same number of 
columns, return the P_AB matrix over the same ring of A and B, where P_AB = 
[[A,A],[0,B]].
}
	require Ncols(A) eq Ncols(B):
		"matrices must have the same number of columns";
	require BaseRing(A) eq BaseRing(B):
		"matrices must have the same base ring";
	ring := BaseRing(A);
	
	row1 := HorizontalJoin(A,A);
	row2 := HorizontalJoin(ZeroMatrix(ring,Nrows(B),Ncols(B)), B);
	P_AB := VerticalJoin(row1,row2);
	
	return P_AB,LinearCode(P_AB);	
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** PlotkinSum function
******************************************************************************/

intrinsic PlotkinSum(C::Code,D::Code) -> Code
{
	Given codes C and D both over the same ring and of the same length, 
construct the Plotkin sum of C and D. The Plotkin sum consists of all vectors 
of the form (u|u+v), for all u in C and v in D.
}
	require Length(C) eq Length(D):
		"codes must have the same length";
	require BaseRing(C) eq BaseRing(D):
		"codes must be defined over the same ring";
	
	Gc := GeneratorMatrix(C);
	Gd := GeneratorMatrix(D);
	G  := PlotkinSum(Gc,Gd); 
	
	return LinearCode(G);		
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** BQPlotkinSum function
******************************************************************************/

intrinsic BQPlotkinSum(A::Mtrx,B::Mtrx,C::Mtrx) -> Mtrx
{
	Given three matrices A, B, and C over Z4, all with the same number of 
columns, return the BQP_ABC matrix over Z4, where BQP_ABC = 
[[A,A,A,A],[0,B',2B',3B'],[0,0,B'',B''],[0,0,0,C]]. The matrix B' is obtained 
from B switching twos by ones in the rows of order two, and B'' is obtained 
from B removing the rows of order two. 
}
	require Ncols(A) eq Ncols(B) and Ncols(A) eq Ncols(C):
		"matrices must have the same number of columns";
	require BaseRing(A) eq BaseRing(B) and BaseRing(A) eq BaseRing(C) 
			and BaseRing(A) eq Z4:
		"matrices must be over Z4";
	
	n  := Ncols(A);
	mb := Nrows(B);
	mc := Nrows(C);
	
	row1 	:= HorizontalJoin(<A,A,A,A>);

	Bprime 	:= Replace2with1(B);
	zero2 	:= ZeroMatrix(Z4, mb, n);
	row2 	:= HorizontalJoin(<zero2,Bprime,2*Bprime,3*Bprime>);

	Bhut 	:= RemoveOrder2Rows(B);
	mbhut 	:= Nrows(Bhut);
	zeroHut := ZeroMatrix(Z4, mbhut, n);

	row3 	:= HorizontalJoin(<zeroHut,zeroHut,Bhut,Bhut>);

	zero3 	:= ZeroMatrix(Z4, mc, n);
	row4 	:= HorizontalJoin(<zero3,zero3,zero3,C>);

	BQP_ABC := VerticalJoin(<row1,row2,row3,row4>);
		
	return BQP_ABC,LinearCode(BQP_ABC);
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** BQPlotkinSum function
******************************************************************************/

intrinsic BQPlotkinSum(C::Code,D::Code,E::Code) -> Code
{
	Given three codes C,D and E over Z4, all the same length, construct the 
BQ-Plotkin sum of C,D and E. Let Gd be the generator matrix of the code D of 
type 2^gamma 4^delta. The code D' over Z4 is obtained from D replacing the twos 
with ones in the gamma rows of order two of Gd, and the code D'' over Z4 is 
obtained from D removing the gamma rows of order two of Gd.
The BQ-Plotkin sum is a code over Z4 that consists of all vectors of the form 
(u,u+v',u+2v'+v'',u+3v'+v''+z), where u in C, v' in D', v'' in D'' and z in E.
}
	require Length(C) eq Length(D) 
		and Length(C) eq Length(E):
		"codes must have the same length";
	require BaseRing(C) eq BaseRing(D) 
		and BaseRing(C) eq BaseRing(E)
		and BaseRing(C) eq Z4:
		"codes must be defined over Z4";
		
		Gc := GeneratorMatrix(C);
		Gd := GeneratorMatrix(D);
		Ge := GeneratorMatrix(E);
		G  := BQPlotkinSum(Gc,Gd,Ge); 
		return LinearCode(G);
end intrinsic;

/*****************************************************************************/
/*****************************************************************************
** QuaternaryPlotkinSum function
******************************************************************************/

intrinsic QuaternaryPlotkinSum(A::Mtrx,B::Mtrx) -> Mtrx
{
	Given matrices A and B both over Z4 and with the same number of 
columns, return the QP_AB matrix over Z4 of A and B, where QP_AB = 
[[A,A,A,A],[0,B,2B,3B]].
}
	require Ncols(A) eq Ncols(B):
		"matrices must have the same number of columns";
	require BaseRing(A) eq BaseRing(B) and BaseRing(A) eq Z4:
		"matrices must be over Z4";
	
	Zero  := ZeroMatrix(Z4,Nrows(B),Ncols(B));
	row1  := HorizontalJoin(<A,A,A,A>);
	row2  := HorizontalJoin(<Zero,B,2*B,3*B>);
	QP_AB := VerticalJoin(row1,row2);
	return QP_AB,LinearCode(QP_AB);		
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** QuaternaryPlotkinSum function
******************************************************************************/

intrinsic QuaternaryPlotkinSum(C::Code,D::Code) -> Code
{
	Given two codes C and D over Z4, both of the same length, construct the 
Quaternary Plotkin sum of C and D. The Quaternary Plotkin sum is a code over Z4 
that consists of all vectors of the form (u,u+v,u+2v,u+3v), where u in C and v 
in D.
}
	require Length(C) eq Length(D):
		"codes must have the same length";
	require BaseRing(C) eq BaseRing(D)
		and BaseRing(C) eq Z4:
		"codes must be defined over Z4";
	
	Gc := GeneratorMatrix(C);
	Gd := GeneratorMatrix(D);
	G  := QuaternaryPlotkinSum(Gc,Gd); 
	return LinearCode(G);		
end intrinsic;

/*****************************************************************************/
/*****************************************************************************
** DoublePlotkinSum function
******************************************************************************/

intrinsic DoublePlotkinSum(A::Mtrx,B::Mtrx,C::Mtrx,D::Mtrx) -> Mtrx
{
	Given matrices A,B,C, and D, all over Z4 and with the same 
number of columns, return the DP_ABCD matrix over Z4, where QP_AB = 
[[A,A,A,A],[0,B,2B,3B],[0,0,C,C],[0,0,0,D]].
}
	require Ncols(A) eq Ncols(B) and 
			Ncols(A) eq Ncols(C) and 
			Ncols(A) eq Ncols(D):
		"matrices must have the same number of columns";
	require BaseRing(A) eq BaseRing(B) and 
			BaseRing(A) eq BaseRing(C) and
			BaseRing(A) eq BaseRing(D) and
			BaseRing(A) eq Z4:
		"matrices must be over Z4";
	
	n := Ncols(A);
	
	ZeroB := ZeroMatrix(Z4,Nrows(B),n);
	ZeroC := ZeroMatrix(Z4,Nrows(C),n);
	ZeroD := ZeroMatrix(Z4,Nrows(D),n);
	
	row1 := HorizontalJoin(<A,A,A,A>);
	row2 := HorizontalJoin(<ZeroB,B,2*B,3*B>);
	row3 := HorizontalJoin(<ZeroC,ZeroC,C,C>);
	row4 := HorizontalJoin(<ZeroD,ZeroD,ZeroD,D>);
	
	DP_ABCD := VerticalJoin(<row1,row2,row3,row4>);	
	return DP_ABCD,LinearCode(DP_ABCD);			
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** DoublePlotkinSum function
******************************************************************************/

intrinsic DoublePlotkinSum(C::Code,D::Code,E::Code,F::Code) -> Code
{
	Given four codes C, D, E, and F over Z4, all of the same length, construct 
the Double Plotkin sum od C, D, E, and F. The Double Plotkin sum is a code over 
Z4 that consists of all vectors of the form (u,u+v,u+2v+z,u+3v+z+t), where u in 
C, v in D, z in E, and t in F.
}
	require Length(C) eq Length(D) 
		and Length(C) eq Length(E)
		and Length(C) eq Length(F):
		"codes must have the same length";
	require BaseRing(C) eq BaseRing(D) 
		and BaseRing(C) eq BaseRing(E)
		and BaseRing(C) eq BaseRing(F)
		and BaseRing(C) eq Z4:
		"codes must be defined over Z4";
		
		Gc := GeneratorMatrix(C);
		Gd := GeneratorMatrix(D);
		Ge := GeneratorMatrix(E);
		Gf := GeneratorMatrix(F);
		G  := DoublePlotkinSum(Gc,Gd,Ge,Gf); 
		return LinearCode(G);
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** ReedMullerCodeRMZ4Matrix function
******************************************************************************/
/*	Description:
	Given integers s,r and m, return a generator matrix for the Reed-Muller
	code over Z4 given by the ReedMullerCodeRMZ4 function
*/
function ReedMullerCodeRMZ4Matrix(s,r,m)
	case m:
	when 1:			
		case r:
			when 0:
				return GeneratorMatrix(RepetitionCodeZ4(2^0));
			else
				if r lt 0 then
					return GeneratorMatrix(ZeroCode(Z4,1));
				else 
					return GeneratorMatrix(UniverseCode(Z4,1));
				end if;
		end case;	
						 
	else //m>1
		if r lt 0 then 
			return GeneratorMatrix(ZeroCode(Z4,2^(m-1)));
		else
			if s lt (Ceiling((m-1)/2)) then
				// Plotkin construction
				return  PlotkinSum(	ReedMullerCodeRMZ4Matrix(s,r,m-1), 
									ReedMullerCodeRMZ4Matrix(s,r-1,m-1));
			else
				// BQ-Plotkin construction
					return RemoveZeroRows(
							BQPlotkinSum(
									ReedMullerCodeRMZ4Matrix(s-1,r  ,m-2), 
									ReedMullerCodeRMZ4Matrix(s-1,r-1,m-2), 
									ReedMullerCodeRMZ4Matrix(s-1,r-2,m-2))
									);
			end if;
		end if;
	end case;
end function;

/*****************************************************************************/
/******************************************************************************
** ReedMullerCodeRMZ4 function
******************************************************************************/ 
intrinsic ReedMullerCodeRMZ4(s::RngIntElt,r::RngIntElt,m::RngIntElt) -> CodeLinRng,Mtrx
{
	 Given an integer m >= 1, an integer r such that 0 <= r <= m, and an 
integer s such that 0 <= s <= floor((m - 1)/2), return a r-th order Reed-Muller 
code over Z4 of length 2^(m-1), denoted by RMs(r,m). The binary image under the 
Gray map is a binary (not necessarily linear) code with the same parameters as 
the binary linear r-th order Reed-Muller code of length 2^m. Note that the 
inclusion and duality properties are also satisfied, that is, the code 
RMs(r-1,m) is a subcode of RMs(r,m), r > 0, and the code RMs(r,m) is the dual 
code of RMs(m-r-1,m), r < m.
}
	requirege	 m,  1;
	requirerange s,  0, (m-1) div 2;
	requirerange r, 0, m;
	G := ReedMullerCodeRMZ4Matrix(s,r,m);
	C := LinearCode(G);
	return C,G;		
end intrinsic;


/******************************************************************************
** ReedMullerCodesRMZ4 function
******************************************************************************/

intrinsic ReedMullerCodesRMZ4(s::RngIntElt,m::RngIntElt) -> Tup
{
	Given an integer m >= 1, and an integer s such that 0 <= s <= floor((m - 
1)/2), return the family of Reed-Muller codes over Z4 of length 2^(m-1) , that 
is, the codes RMs(r,m), for all 0 <= r <= m. The binary image of these codes 
under the Gray map gives a family of binary (not necessarily linear) codes with 
the same parameters as the binary linear Reed-Muller family of codes of length 
2m. Note that RMs(0,m) subcode RMs(1,m) subcode ... subcode RMs(m,m).
}
	requirege	 m,  1;
	requirerange s,  0, (m-1) div 2;
	return [ReedMullerCodeRMZ4(s,r,m):r in [0..m]];	
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** ExtendedPerfectCodeZ4 function
******************************************************************************/

intrinsic ExtendedPerfectCodeZ4 
	(delta::RngIntElt,m::RngIntElt) -> CodeLinRng,Mtrx
{
	Given an integer m >= 2 and an integer delta such that 1 <= delta <=
floor((m + 1)/2), return an extended perfect code over Z4 of length 2^(m-1) , 
such that its dual code is of type 2^gamma 4^delta , where gamma = m + 1 - 
2delta. Moreover, return a generator matrix constructed in a recursive way from 
the Plotkin and BQPlotkin constructions deﬁned in Section 2.3 of the manual. An 
extended perfect code over Z4 of length 2^(m−1) is a code over Z4 such that, 
after the Gray map, give a binary (not necessarily linear) code with the same 
parameters as the binary extended perfect code of length 2^m. 

}
	requirege	 m,  2;
	requirerange delta,  1, (m+1) div 2;
	s := delta-1; 
	return ReedMullerCodeRMZ4(s,m-2,m);
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** HadamardCodeZ4 function
******************************************************************************/

intrinsic HadamardCodeZ4(delta::RngIntElt,m::RngIntElt) -> CodeLinRng,Mtrx
{
	Given an integer m >= 2 and an integer δ such that 1 <= delta <= floor((m + 
1)/2), return a Hadamard code over Z4 of length 2^(m-1) and type 2^gamma 
4^delta, where gamma = m + 1 − 2delta. Moreover, return a generator matrix with 
gamma + delta rows constructed in a recursive way from the Plotkin and 
BQPlotkin constructions deﬁned in Section 2.3 of the manual. A Hadamard code 
over Z4 of length 2^(m-1)  is a code over Z4 such that, after the Gray map, 
give a binary (not necessarily linear) code with the same parameters as the 
binary Hadamard code of length 2^m. 
}
	requirege	 m,  2;
	requirerange delta,  1, (m+1) div 2;
	s := delta-1;
	return ReedMullerCodeRMZ4(s,1,m);	
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** BilinearDiagonal function
******************************************************************************/
/*	Description:
	Return the vector (1,3,3,1,3,1,1,3, ...) of length 2^n over Z4.
*/
function BilinearDiagonal(n)
	case n:
		when 0:
			return [Z4!1];
		else
  			return 	BilinearDiagonal(n-1) cat 
					[x+2:x in BilinearDiagonal(n-1)];
	end case;
end function;

/******************************************************************************
** DualKroneckerZ4 function
******************************************************************************/

intrinsic DualKroneckerZ4(C::CodeLinRng) -> CodeLinRng
{
	Given a code C over Z4 of length 2^m, return its Kronecker dual code. 4 The  
Kronecker dual code of C is Cdk = \{x in Z4^(2^m) 4 : x K2^m y = 0, for all y 
in C \}. Equivalently, K2^m is a quaternary matrix of length 2^m with the vector (1,3,3,1,3,1,1,3, ...) in the main diagonal and zeros elsewhere. 
}
	require BaseRing(C) eq Z4:
		"code must be over Z4";
		
	CDual := Dual(C);
	Gd := GeneratorMatrix(CDual);
	m := Nrows(Gd);
	n := Ncols(Gd);
	K := BilinearDiagonal(Valuation(n,2));
	
	for j:=1 to n do
		if(K[j] ne 1) then
			for i:=1 to m do
				Gd[i][j] := K[j]*Gd[i][j];
			end for;
		end if;
	end for;

	return LinearCode(Gd);		
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** DimensionOfSpanZ2 function
******************************************************************************/

intrinsic DimensionOfSpanZ2(C::CodeLinRng) -> RngIntElt
{
	Given a code C over Z4, return the dimension of the linear span of Cbin, 
that is, the dimension of <Cbin>, where Cbin = phi(C) and phi is the Gray map.
}
	require BaseRing(C) eq Z4:
		"code must be over Z4";
		
	GS          		:= GeneratorMatrix(StandardForm(C));
	GMin,gamma,delta    := MinRowsGeneratorMatrix(C); 
	n := Ncols(GS);
	S := RSpace(Z2,n-delta);

	AB := Submatrix(GS, 1, delta+1, delta, n-delta);
	IgC := Submatrix(GS, delta+1, delta+1, gamma, n-delta);
	IgC := Parent(IgC) ! [IntegerRing()!x div 2: x in Eltseq(IgC)];

	generatorsOverZ2 := {S!IgC[i] : i in [1..gamma]};
	for i in [1..delta] do
		for j in [i+1..delta] do
			Include(~generatorsOverZ2,S!AB[i]*S!AB[j]);
		end for;
	end for;
	
	return Dimension(sub<S|generatorsOverZ2>)+2*delta;
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** RankZ2 function
******************************************************************************/

intrinsic RankZ2(C::CodeLinRng) -> RngIntElt
{
	Given a code C over Z4, return the dimension of the linear span of Cbin, 
that is, the dimension of <Cbin>, where Cbin = phi(C) and phi is the Gray map.
}
	require BaseRing(C) eq Z4:
		"code must be over Z4";
		
	return DimensionOfSpanZ2(C);
end intrinsic;

/*****************************************************************************/
/*****************************************************************************
** SpanZ2CodeZ4 function
******************************************************************************/

intrinsic SpanZ2CodeZ4(C::CodeLinRng) -> CodeLinFld
{
	Given a code C over Z4 of length n, return Sc = phi^(−1)(Sbin) as a code 
over Z4, and the linear span of Cbin , Sbin = ⟨Cbin⟩, as a binary linear code 
of length 2^n, where Cbin = phi(C) and phi is the Gray map.
}
	require BaseRing(C) eq Z4:
		"code must be over Z4";
		
	GMin,gamma,delta    := MinRowsGeneratorMatrix(C);
	n 					:= Ncols(GMin);
	Ggamma 				:= Submatrix(GMin, 1, 1, gamma, n);
	Gdelta 				:= Submatrix(GMin, gamma+1, 1, delta, n);
	newVectors  		:= [];
	
	for i:=1 to delta do
		for j:=i to delta do
			newVectors := newVectors cat [2*Gdelta[i]*Gdelta[j]];
		end for;
	end for;
	
	Gnew 		:= Matrix(newVectors);
	GSpanCodeZ4 := VerticalJoin(<Ggamma,Gdelta,Gnew>);
	SpanCodeZ4  := LinearCode(GSpanCodeZ4);
	_,SpanCodeZ2 := HasLinearGrayMapImage(SpanCodeZ4);
	
	return SpanCodeZ4,SpanCodeZ2;		
end intrinsic;

/*****************************************************************************/
/******************************************************************************
** isInKernel function
******************************************************************************/

isInKernel := function(v,Order4Mult2,Order2InKernel)
	isIn := true;
	for z in Order4Mult2 do
		vz := v * z;
		if (vz notin Order2InKernel) then
			isIn := false;
			break;
		end if;
	end for;
	return isIn;
end function;


/*****************************************************************************/
// *****************************************************************************
// ** KernelZ2CodeZ4 function
// ******************************************************************************/
// 
// intrinsic KernelZ2CodeZ4(C::CodeLinRng) -> CodeLinRng
// {
// 	Given a code C over Z4 of length n, return its kernel KC as a subcode over 
// Z4 of C , and Kbin = phi(KC) as a binary linear subcode of Cbin of length 2n, 
// where Cbin = phi(C) and phi is the Gray map. The kernel KC contains the 
// codewords v such that 2v*u in C for all u in C , where ∗ denotes the 
// component-wise product. Equivalently, the kernel Kbin = phi(KC) contains the 
// codewords c in Cbin such that c + Cbin = Cbin , where Cbin = phi(C) and phi is 
// the Gray map. 	
// }
// 
// 	require BaseRing(C) eq Z4:
// 		"code must be over Z4";
// 		
// 	G,gamma,delta 	:= MinRowsGeneratorMatrix(C);
// 	n 				:= Ncols(G);
// 	
// 	// gamma + delta generators of order two in the kernel
// 	Order4Mult2 := [2*G[v]: v in [gamma+1..Nrows(G)]];
// 	Order2InKernel := LinearCode<Z4,n|[G[u]: u in [1..gamma]] 
// 													cat Order4Mult2>;
// 	Order4InKernel := [];
// 		
// 	// generators of order four to test whether are in the kernel or not
// 	Order4NotInKernel := RowSubmatrix(G,gamma+1,delta);
// 	i := Nrows(Order4NotInKernel);
// 
// 	while (i ge 1) do
// 		//the number of vectors to test this iteration
// 		num := Nrows(Order4NotInKernel) - i;
// 
// 		/* CombSpace is a binary space used to generate all the linear 
// 		   combinations of num elements. */
// 		CombSpace := VectorSpace(Z2,num);
// 
// 		if (Dimension(CombSpace) eq 0) then
// 			v := Order4NotInKernel[i];
// 			// if v is in the kernel of C
// 			if (isInKernel(v,Order4Mult2,Order2InKernel) eq true) then
// 				RemoveRow(~Order4NotInKernel,i);
// 				Append(~Order4InKernel,v);
// 			end if;
// 		else
// 			Mcomb := RowSubmatrix(Order4NotInKernel,i+1,num);
// 
// 			// Test all possible combinations (Complexity bottleneck)
// 			for Vcomb in CombSpace do
// 				nonKernelComb := Vcomb * Mcomb;
// 				v := Order4NotInKernel[i] + nonKernelComb;
// 				// if v is in the Kernel of C
// 				if (isInKernel(v,Order4Mult2,Order2InKernel) eq true) then
// 					RemoveRow(~Order4NotInKernel,i);
// 					Append(~Order4InKernel,v);
// 					break;
// 				end if;
// 			end for;
// 		end if;
// 		i := i - 1;
// 	end while;
// 
// 	//kernel over Z4
// 	KernelCodeZ4 := Order2InKernel + LinearCode<Z4,n|Order4InKernel>; 
// 	//generator matrix of C containing the kernel generators
// 	Ggamma := RowSubmatrix(G,1,gamma);
// 	Order4InKernel := Matrix(Z4,#Order4InKernel,n,Order4InKernel);
// 	NewG := VerticalJoin(<Ggamma,Order4InKernel,Order4NotInKernel>);
// 	//kernel over Z2
// 	_,KernelCodeZ2 := HasLinearGrayMapImage(KernelCodeZ4);
// 	
// 	return KernelCodeZ4,KernelCodeZ2,NewG;
// end intrinsic;
// 
// 
// /****************************************************************************
/******************************************************************************
** DimensionOfKernelZ2 function
******************************************************************************/

intrinsic DimensionOfKernelZ2(C::CodeLinRng) -> RngIntElt
{
	Given a code C over Z4, return the dimension of the Gray map image of its 
kernel KC over Z4, that is the dimension of Kbin = phi(KC), where phi is the 
Gray map. Note that Kbin is always a binary linear code.
}
	require BaseRing(C) eq Z4:
		"code must be over Z4";
		
	_,KernelCodeZ2 := KernelZ2CodeZ4(C);
	return Dimension(KernelCodeZ2);		
end intrinsic;

/*****************************************************************************/
/*****************************************************************************/
// Auxiliar functions for LRM codes
/*****************************************************************************/

ZipSets := func < f,A,B| {f(x,y): x in A, y in B}>;

ZipTups := func < f,A,B| <f(x,y): x in A, y in B>>;

function JoinTups(t1,t2)
	case #t2:
	when 0: return t1;
	else 
		 return JoinTups(Append(t1,t2[#t2]),Prune(t2));
	end case;
end function;

/*****************************************************************************/
/******************************************************************************
** LRM function
******************************************************************************/
/*	Description:
	Return the set of LRM codes for r and m parameters.
*/
function LRM(r,m)

	nh := (m+1) div 2; //delta for the new hadamard in case m.
	case r:
		when -1: /* Zero code */
			return <GeneratorMatrix(ZeroCode(Z4,2^(m-1)))>;

		when 0:  /* Repetition code */
			return <GeneratorMatrix(RepetitionCodeZ4(2^(m-1)))>;
		

		when m:  /* Universe code */ 
			return  <GeneratorMatrix(UniverseCode(Z4,2^(m-1)))>;

		else  
		    newHadamard := <>;
			if(r eq 1 and m ge 3 and m mod 2 ne 0) then //hadamard case
				_,G := HadamardCodeZ4(nh,m);
				Append(~newHadamard,G);
			end if;
			return JoinTups(
					ZipTups(
						PlotkinSum,LRM(r, m-1),LRM(r-1, m-1)),newHadamard);
	end case;
end function;

/*****************************************************************************/
/******************************************************************************
** ReedMullerCodesLRMZ4 function
******************************************************************************/

intrinsic ReedMullerCodesLRMZ4(r::RngIntElt,m::RngIntElt) -> SeqEnum
{
	Given an integer m >= 1 and an integer r such that 0 <= r <= m, return a
set of r-th order Reed-Muller codes over Z4 of length 2^(m-1). The binary image 
under the Gray map of any of these codes is a binary (not necessarily linear) 
code with the same parameters as the binary linear r-th order Reed-Muller code 
of length 2^m. Note that for these codes neither the usual inclusion nor 
duality properties of the binary linear Reed-Muller family are satisﬁed. 
}
	requirege 		m,  1;
	requirerange 	r,  0, m;
	tupla := LRM(r,m);
	return {LinearCode(G):G in tupla};
end intrinsic;

/*****************************************************************************/
/*****************************************************************************
** KernelZ2CodeZ4 function
******************************************************************************/
intrinsic KernelZ2CodeZ4(C::CodeLinRng) -> CodeLinRng
{
	Given a code C over Z4 of length n, return its kernel KC as a subcode over 
Z4 of C , and Kbin = phi(KC) as a binary linear subcode of Cbin of length 2n, 
where Cbin = phi(C) and phi is the Gray map. The kernel KC contains the 
codewords v such that 2v*u in C for all u in C , where ∗ denotes the 
component-wise product. Equivalently, the kernel Kbin = phi(KC) contains the 
codewords c in Cbin such that c + Cbin = Cbin , where Cbin = phi(C) and phi is 
the Gray map.	
}
	require BaseRing(C) eq Z4:
		"code must be over Z4";
		
	isLinear,BinaryLinearCode := HasLinearGrayMapImage(C);	
	
	if isLinear then
		return C,BinaryLinearCode;
	else
		G   := GeneratorMatrix(C);
		G4  := [G[i] : i in [1..Nrows(G)] | not IsZero(G[i] + G[i]) ];
		H   := GeneratorMatrix(Dual(C));
		H4  := [H[i] : i in [1..Nrows(H)] | not IsZero(H[i] + H[i]) ];
		GH4 := [2*u*v : u in G4, v in H4];
		HH  := VerticalJoin(H,Matrix(GH4));
		  KernelCodeZ4 := Dual(LinearCode(HH));
		_,KernelCodeZ2 := HasLinearGrayMapImage(KernelCodeZ4);
			return KernelCodeZ4,KernelCodeZ2;
	end if;
	
end intrinsic;

