
freeze;

declare type HypGeomData;
declare attributes HypGeomData: alpha,beta,alpha_poly,beta_poly;
declare attributes HypGeomData: AN,AD,BN,BD,cycA,cycB,Mvalue,dvalue,fps;
declare attributes HypGeomData: bad,char,Garray,Glist;
declare attributes HypGeomData: hodge,weight;

Z:=Integers(); Q:=Rationals();

intrinsic Print(X::HypGeomData,level::MonStgElt) {}
 printf "Hypergeometric data given by %o and %o",X`alpha,X`beta; end intrinsic;

intrinsic Weight(X::HypGeomData) -> RngIntElt
{Return the weight of given hypergeometric data}
 return X`weight; end intrinsic;

intrinsic Degree(X::HypGeomData) -> RngIntElt
{Return the degree of given hypergeometric data}
 return #X`alpha; end intrinsic;

intrinsic DefiningPolynomials(X::HypGeomData) -> RngUPolElt, RngUPolElt
{The polynomials (products of cyclotomics) that define X}
 return X`alpha_poly,X`beta_poly; end intrinsic;

intrinsic CyclotomicData(X::HypGeomData) -> SeqEnum, SeqEnum
{The cyclotomic data for X}
 return X`cycA,X`cycB; end intrinsic;

intrinsic AlphaBetaData(X::HypGeomData) -> SeqEnum, SeqEnum
{The alpha-beta data for X}
 return X`alpha,X`beta; end intrinsic;

intrinsic MValue(X::HypGeomData) -> FldRatElt
{The scaling parameter M associated to X}
 return X`Mvalue; end intrinsic;

intrinsic GammaList(X::HypGeomData) -> List
{The list of gamma corresponding to X}
 return X`Glist; end intrinsic;

intrinsic GammaArray(X::HypGeomData) -> SeqEnum
{The array of gamma corresponding to X}
 return X`Garray; end intrinsic;

intrinsic IsPrimitive(X::HypGeomData) -> BoolElt, RngIntElt
{Whether the given hypergeometric data is primitive}
 g:=Gcd([x : x in X`Glist]); return g eq 1,g; end intrinsic;

intrinsic PrimitiveData(X::HypGeomData) -> HypGeomData
{Return the primitivization of X}
 b,g:=IsPrimitive(X); if b then return X; end if;
 return HypergeometricData([* x div g : x in X`Glist *]); end intrinsic;

intrinsic 'eq'(X::HypGeomData,Y::HypGeomData) -> BoolElt
{Whether two instances of hypergeometric data are equal}
 return X`alpha eq Y`alpha and X`beta eq Y`beta; end intrinsic;

intrinsic 'ne'(X::HypGeomData,Y::HypGeomData) -> BoolElt
{Whether two instances of hypergeometric data are equal}
 return not (X eq Y); end intrinsic;

////////////////////////////////////////////////////////////////////////

function getGlist(G) X:=[];
 for i in [1..#G] do X:=X cat[Sign(G[i])*i : j in [1..Abs(G[i])]]; end for;
 return SequenceToList(X); end function;

function hgm_init(ALPHA,BETA,AD,BD)
 X:=New(HypGeomData); X`alpha:=ALPHA; X`beta:=BETA;
 X`AN:=[Numerator(x) : x in ALPHA]; X`AD:=AD;
 X`BN:=[Numerator(x) : x in BETA]; X`BD:=BD; X`bad:=LCM(X`AD cat X`BD);
 S:=[<x,0> : x in ALPHA] cat [<1-x,1> : x in BETA]; Sort(~S);
 s:=0; t:=FunctionField(Rationals()).1; f:=Parent(t)!0;
 for i in [1..2*#ALPHA] do
  if S[i][2] eq 0 then f:=f+t^s; s:=s+1; else s:=s-1; end if; end for;
 X`hodge:=f;
 X`weight:=Degree(Numerator(X`hodge)); X`dvalue:=Degree(Denominator(X`hodge));
 n:=LCM(X`AD); K:=CyclotomicField(n); zn:=K.1; P:=PolynomialRing(K); y:=P.1;
 gn:=ChangeRing(&*[y-zn^(Integers()!(n*ALPHA[i])) : i in [1..#ALPHA]],Q);
 d:=LCM(X`BD); K:=CyclotomicField(d); zd:=K.1; P:=PolynomialRing(K); y:=P.1;
 gd:=ChangeRing(&*[y-zd^(Integers()!(d*BETA[i])) : i in [1..#ALPHA]],Q);
 X`alpha_poly:=gn; X`beta_poly:=gd; X`cycA:=[]; X`cycB:=[];
 FA:=Factorization(gn); FB:=Factorization(gd); G:=[0:i in [1..LCM(n,d)]];
 for x in FA do
  _,v:=IsCyclotomicPolynomial(x[1]); X`cycA cat:=[v : i in [1..x[2]]];
  for d in Divisors(v) do u:=v div d; G[d]:=G[d]+x[2]*MoebiusMu(u);
   end for; end for;
 for x in FB do
  _,v:=IsCyclotomicPolynomial(x[1]); X`cycB cat:=[v : i in [1..x[2]]];
  for d in Divisors(v) do u:=v div d; G[d]:=G[d]-x[2]*MoebiusMu(u);
   end for; end for;
 assert &+[G[i]*i : i in [1..#G]] eq 0;
 X`cycA:=Sort(X`cycA); X`cycB:=Sort(X`cycB);
 X`Garray:=G[1..Max([d : d in [1..#G] | G[d] ne 0])];
 X`Glist:=getGlist(X`Garray);
 spin:=func<N|&*[d^(d*MoebiusMu(N div d)) : d in Divisors(N)]>;
 X`Mvalue:=&*[spin(x) : x in X`cycA]/&*[spin(y) : y in X`cycB];
 X`fps:=[FractionalPart(-x) : x in X`beta];
 P:=PolynomialRing(Rationals()); X`char:=P!1; T:=P.1;
 if X`weight mod 2 eq 0 then
  ARR:=IsCoercible(Integers(),&+ALPHA) select X`cycA else X`cycB;
  X`char:=(-T)^(#ALPHA mod 2)*(1-T)*
          &*[Discriminant(CyclotomicPolynomial(d)) : d in ARR]; end if;
 return X; end function;

intrinsic Twist(X::HypGeomData) -> HypGeomData
{The twist of X determined by shifting all alpha and beta by 1/2}
 A:=[FractionalPart(x+1/2) : x in X`alpha];
 B:=[FractionalPart(x+1/2) : x in X`beta]; Sort(~A); Sort(~B);
 if B[1] lt A[1] then return HypergeometricData(A,B);
 else return HypergeometricData(B,A); end if; end intrinsic;

intrinsic HypergeometricData(f::RngUPolElt,g::RngUPolElt) -> HypGeomData
{The hypergeometric data given by cyclomtomic polynomial products}
 require Degree(f) eq Degree(g): "Degrees must be equal";
 if BaseRing(Parent(f)) eq Z then f:=ChangeRing(f,Q); end if;
 if BaseRing(Parent(g)) eq Z then g:=ChangeRing(g,Q); end if;
 require BaseRing(Parent(f)) eq Q: "f must be defined over Q";
 require BaseRing(Parent(g)) eq Q: "g must be defined over Q";
 require Degree(Gcd(f,g)) eq 0: "Polynomials must be coprime";
 F:=Factorization(f); ALPHA:=[];
 require &and[IsCyclotomicPolynomial(h[1]) : h in F]: "f must be cyclotomic";
 for h in F do for j in [1..h[2]] do _,d:=IsCyclotomicPolynomial(h[1]);
  ALPHA cat:=[Q!(x/d) : x in [1..d] | Gcd(x,d) eq 1]; end for; end for;
 G:=Factorization(g); BETA:=[];
 require &and[IsCyclotomicPolynomial(h[1]) : h in G]: "g must be cyclotomic";
 for h in G do for j in [1..h[2]] do _,d:=IsCyclotomicPolynomial(h[1]);
  BETA cat:=[Q!(x/d) : x in [0..d-1] | Gcd(x,d) eq 1]; end for; end for;
 return HypergeometricData(ALPHA,BETA); end intrinsic;

intrinsic
HypergeometricData(F::SeqEnum[RngIntElt],G::SeqEnum[RngIntElt]) -> HypGeomData
{The hypergeometric data given by integer sequences corresponding to
 cyclotomic polynomials}
 require &and[x gt 0 : x in F cat G]: "Integers must be positive";
 require #F ne 0 and #G ne 0: "Sequences must be non-empty";
 require #(Set(F) meet Set(G)) eq 0: "Sequences must be disjoint";
 F:=&*[CyclotomicPolynomial(f) : f in F];
 G:=&*[CyclotomicPolynomial(g) : g in G];
 require Degree(F) eq Degree(G): "Sequences must define same degree";
 return HypergeometricData(F,G); end intrinsic;

intrinsic HypergeometricData(ALPHA::SeqEnum,BETA::SeqEnum) -> HypGeomData
{The hypergeometric data given by alphas and betas (rationals)}
 if Universe(ALPHA) eq Z then ALPHA:=ChangeUniverse(ALPHA,Q); end if;
 if Universe(BETA) eq Z then BETA:=ChangeUniverse(BETA,Q); end if;
 require Universe(ALPHA) eq Rationals(): "ALPHA must be in rationals";
 require Universe(BETA) eq Rationals(): "BETA must be in rationals";
 require #ALPHA eq #BETA: "ALPHA and BETA must be same size";
 require #(Set(ALPHA) meet Set(BETA)) eq 0: "ALPHA and BETA cannot intersect";
 ALPHA:=Sort([(Numerator(x) mod Denominator(x))/Denominator(x) : x in ALPHA]);
 BETA:=Sort([(Numerator(x) mod Denominator(x))/Denominator(x) : x in BETA]);
 if BETA[1] gt ALPHA[1] then T:=BETA; BETA:=ALPHA; ALPHA:=T;
  print "WARNING: Switching ALPHA/BETA (latter now has least element)"; end if;
 AD:=[Denominator(x) : x in ALPHA]; BD:=[Denominator(x) : x in BETA];
 for d in AD do if d eq 1 or d eq 2 then continue; end if;
  MS:=Multiset([Numerator(x) : x in ALPHA | Denominator(x) eq d]); S:=Set(MS);
  require #S eq EulerPhi(d): "ALPHA must have all conjugates mod",d;
  m:=Multiplicity(MS,1);
  for s in S do require Multiplicity(MS,s) eq m:
	          "ALPHA has unequal conjugates mod",d; end for; end for;
 for d in BD do if d eq 1 or d eq 2 then continue; end if;
  MS:=Multiset([Numerator(x) : x in BETA | Denominator(x) eq d]); S:=Set(MS);
  require #S eq EulerPhi(d): "BETA must have all conjugates mod",d;
  m:=Multiplicity(MS,1);
  for s in S do require Multiplicity(MS,s) eq m:
	          "BETA has unequal conjugates mod",d; end for; end for;
 return hgm_init(ALPHA,BETA,AD,BD); end intrinsic;

intrinsic HypergeometricData(GAMMA::SeqEnum[RngIntElt]) -> HypGeomData
{The hypergeometric data given by an array of gammas}
 require #GAMMA gt 0: "GAMMA must not be empty";
 require &+[GAMMA[x]*x : x in [1..#GAMMA]] eq 0: "sum xG[x] must be 0";
 T:=FunctionField(Rationals()).1;
 prod:=&*[(T^v-1)^GAMMA[v] : v in [1..#GAMMA] | GAMMA[v] ne 0];
 return HypergeometricData(Numerator(prod),Denominator(prod)); end intrinsic;

intrinsic HypergeometricData(GAMMA::List) -> HypGeomData
{The hypergeometric data given by a list of gammas}
 require #GAMMA gt 0: "GAMMA must not be empty";
 require &and[IsCoercible(Z,x) : x in GAMMA]: "GAMMA must have integers";
 GAMMA:=[Z!x : x in GAMMA]; require &+GAMMA eq 0: "Sum must be 0";
 require not 0 in GAMMA: "GAMMA cannot have 0 in it";
 m:=Max([Abs(x) : x in GAMMA]); G:=[0 : i in [1..m]];
 for u in GAMMA do G[Abs(u)]:=G[Abs(u)]+Sign(u); end for;
 return HypergeometricData(G); end intrinsic;

intrinsic HypergeometricData(INPUT::SeqEnum[SeqEnum]) -> HypGeomData
{The hypergeometric derived from an arrays of 2 arrays}
 require #INPUT eq 2: "Must have two arrays";
 return HypergeometricData(INPUT[1],INPUT[2]); end intrinsic;

////////////////////////////////////////////////////////////////////////

function coprime_fracs(n)
 return [k/n : k in [0..n-1] | Gcd(k,n) eq 1]; end function;

function direct_weight(ALPHA,BETA)
 S:=[<x,0> : x in ALPHA] cat [<1-x,1> : x in BETA]; Sort(~S);
 s:=0; t:=FunctionField(Rationals()).1; f:=Parent(t)!0;
 for i in [1..2*#ALPHA] do
  if S[i][2] eq 0 then f:=f+t^s; s:=s+1; else s:=s-1; end if; end for;
 return Degree(Numerator(f)); end function;

function mysort(A)
 f:=func<x,y|(x[2] lt y[2] or (x[2] eq y[2] and x[1] lt y[1]))
 	      select -1 else 1>; B:=Sort(A,f); return B; end function;

function IsTwistMinimal(x) y:=[[FractionalPart(u+1/2) : u in a] : a in x];
 xd:=Reverse(Sort([Denominator(a) : a in x[1] cat x[2]]));
 yd:=Reverse(Sort([Denominator(a) : a in y[1] cat y[2]]));
 if xd lt yd then return true; end if; if xd gt yd then return false; end if;
 x:=[Sort(u) : u in x]; y:=[Sort(u) : u in y];
 if x[1] lt x[2] then x:=[x[2],x[1]]; end if;
 if y[1] lt y[2] then y:=[y[2],y[1]]; end if; return x le y; end function;

function to_cyclo(X) PHI:=[]; A:=X;
 while #A ne 0 do a:=A[1]; m:=Multiplicity(A,a); d:=Denominator(a);
  PHI cat:=[d : i in [1..m]];
  for i in [1..m],x in coprime_fracs(d) do Exclude(~A,x); end for; end while;
 return Sort(PHI); end function;

function primitivity_index(A,B)
 if Universe(A) ne Z then A:=to_cyclo(A); B:=to_cyclo(B); end if;
 G:=[0 : i in [1..Max(A cat B)]];
 for v in A,d in Divisors(v) do G[d]:=G[d]+MoebiusMu(v div d); end for;
 for v in B,d in Divisors(v) do G[d]:=G[d]-MoebiusMu(v div d); end for;
 g:=Gcd([x : x in getGlist(G)]); return g; end function;

intrinsic PossibleHypergeometricData
(d::RngIntElt : Weight:=false,TwistMinimal:=false,
                CyclotomicData:=false,Primitive:=0) -> SeqEnum
{Given a degree d (and weight w), generate the possible hypergeometric data}
 require d ge 1: "Degree must be positive"; w:=Weight;
 require Weight cmpeq false or (Type(w) eq RngIntElt and w ge 0 and w lt d):
 "Weight must be a nonnegative integer and less than degree";
 pr:=Primitive; if Type(pr) eq BoolElt then pr:=pr select 1 else 0; end if;
 require Type(pr) eq RngIntElt and pr ge 0: "Primitivity must be integer >=0";
 if IsOdd(d) and Type(w) eq RngIntElt and IsOdd(w) then return []; end if;
 if pr gt 1 and d mod pr ne 0 then return []; end if;
 // for pr>1, should generate prim data for degree d/pr, and then enlarge it
 E:=[EulerPhiInverse(i) : i in [1..d]]; A:={}; R:=[];
 for P in Partitions(d) do if &or[#E[p] eq 0 : p in P] then continue; end if;
  C:=CartesianProduct([Integers(#E[p]) : p in P]);
  for c in C do
   A join:={Sort(&cat[coprime_fracs(E[P[i]][1+Z!c[i]]) : i in [1..#c]])};
    end for; end for;
 for a in A,b in A do if b[1] ge a[1] then continue; end if;
  if #(Set(a) meet Set(b)) ne 0 then continue; end if; R cat:=[[a,b]]; end for;
 if Type(w) eq BoolElt then X:=mysort(R);
 else X:=mysort([r : r in R | direct_weight(r[1],r[2]) eq w]); end if;
 if TwistMinimal then X:=[x : x in X | IsTwistMinimal(x)]; end if;
 if CyclotomicData then X:=[[to_cyclo(x[1]),to_cyclo(x[2])] : x in X]; end if;
 if pr ne 0 then X:=[x : x in X | primitivity_index(x[1],x[2]) eq pr]; end if;
 return X; end intrinsic;
