freeze;

///////////////////////////////////////////////////////////////////////
// ambient.m
//	products
//	things at infinity
//	nice affine patches and affine decomposition
// April 2001 GDB
//
// October 2002 Nils Bruin
//   - changes to HyperplaneAtInfinity, LineAtInfinity.
//   really weird closure maps installed with MakeProjectiveClosureMap
//   will not work.
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////

intrinsic NumberOfCoordinates(X::Sch) -> RngIntElt
    {The number of coordinates of a point of X.}
    return Length(X);
end intrinsic;

///////////////////////////////////////////////////////////////////////

intrinsic DirectProduct(A::Sch,B::Sch) -> Sch,SeqEnum
{The product of the ambient spaces of A and B together with the two projection
maps}
    require IsAmbient(A) and IsAmbient(B) : 
	    "Arguments must be ambient spaces";
    k := BaseRing(A);
    require BaseRing(B) cmpeq k:
        "Arguments must be defined over the same base ring";
    if (IsAffine(A) or IsOrdinaryProjective(A)) and
	   (IsAffine(B) or IsOrdinaryProjective(B)) then
        dA := Length(A);	// dimension if affine, dimension+1 if projective.
        dB := Length(B);
        if IsAffine(A) and IsAffine(B) then
        P := AffineSpace(k,dA + dB);
        elif IsAffine(A) then
        assert IsOrdinaryProjective(B);
        P := ProjectiveSpace(k,[ 0 : i in [1..dA]] cat [ 1 : i in [1..dB]]);
        elif IsAffine(B) then
        P := ProjectiveSpace(k,[ 1 : i in [1..dA]] cat [ 0 : i in [1..dB]]);
        else 
        P := ProductProjectiveSpace(k,[dA-1,dB-1]);
        end if;
        fPA := map< P -> A | [P.j : j in [1..dA] ] >;
        fPB := map< P -> B | [P.j : j in [dA+1..dA+dB] ] >;
        // Assign the products
        C:=CoxRing(P);
        C`summands:=[PowerStructure(TorVar) | A,B];
        // Return the data
        return P,[fPA,fPB];
    else
        // We can do more in the toric case
        require ISA(Type(A),TorVar) and ISA(Type(B),TorVar):
            "Arguments must be toric varieties";
        require IsField(BaseRing(A)): "The base ring must be a field";
        P,projs:=DirectProduct([PowerStructure(TorVar) | A,B]);
        return P,projs;
    end if;
end intrinsic;

intrinsic DirectProduct(k::Rng,N::SeqEnum) -> Sch,SeqEnum
{The product of projective spaces whose dimensions are the entries of N}
    require &and[ n gt 0 : n in N ]:
	    "The second argument must comprise strictly positive integers";
    return ProductProjectiveSpace(k,N);
end intrinsic;

intrinsic HyperplaneAtInfinity(X::Sch) -> Sch
{The hyperplane at infinity as a subscheme of the projective closure of X}
    require Type(X) eq Aff:
        "Only know about hyperplanes at infinity for affine spaces";

    PX := ProjectiveClosure(X);
    require #Gradings(PX) eq 1:
        "Projective closure is not a projective space";

    assert exists(ind){i:i in [1..#l]|l[i] eq 1} where
               l:=DefiningEquations(ProjectiveClosureMap(X));

    return Scheme(PX,PX.ind);
end intrinsic;
 
intrinsic LineAtInfinity(A::Aff) -> Crv
{The line at infinity as a curve in the projective closure of A}
    require Dimension(A) eq 2:
        "The argument must be an affine plane";
    P := ProjectiveClosure(A);
    require #Gradings(P) eq 1:
        "Projective closure is not a projective plane";

    assert exists(ind){i:i in [1..#l]|l[i] eq 1} where
               l:=DefiningEquations(ProjectiveClosureMap(A));

    return Curve(P,P.ind);
end intrinsic;

intrinsic NumberOfAffinePatches(S::Sch) -> RngIntElt
{The number of standard affine patches of S, returning 0 for an affine scheme}
    A := Ambient(S);
    if Type(A) eq Aff then
        return 0;
    else
	return &*Lengths(A);
    end if;
end intrinsic;

intrinsic AffineDecomposition(P::Prj) -> SeqEnum,Pt
{A sequence of maps from affine spaces whose images partition P together
with the returned point}
    require IsOrdinaryProjective(P):
        "Argument must be ordinary projective space";
    n := Dimension(P);
    k := BaseRing(P);
    spaces := [ AffinePatch(P,1) ] cat
		[ AffineSpace(k,i) : i in [n-1..1 by -1] ];
    maps := [ PCMap(spaces[1]) ];
    for i in [2..n] do
        Ai := spaces[i];
        one := CoordinateRing(Ai) ! 1;
        zero := CoordinateRing(Ai) ! 0;
        m := map<Ai -> P | Fi > where Fi is
	    [ Ai.j : j in [1..n+1-i]] cat [one] cat [zero: j in [1..i-1]];
	maps := maps cat [m];
    end for;
    pt := P ! ([ k!1 ] cat [ 0 : i in [1..n]]);
    return maps,pt;
end intrinsic;

intrinsic CentredAffinePatch(C::Sch,p::Pt) -> Sch,MapSch
{An affine patch of C centred at the point p and the embedding into C}
    require IsProjective(C): "First argument must be projective";
    require p in C(BaseRing(C)):
	"Second argument must be rational point on the first";
    iscurve := ISA(Type(C),Crv);
    Ci,q := AffinePatch(C,p);
    Ai := Ambient(Ci);
    fi := ProjectiveClosureMap(Ai);
    if q eq Origin(Ai) then
        return Ci,fi;
    end if;
    phi := Translation(Ci,q);
    D := Ci @ phi;
    prechart := Inverse(phi) * fi;
    // create a new affine space to enable the correct projective closure map.
    // note that this is a _different_ space from Ai even though they have the
    // same coordinate ring. in particular, they will have different projective
    // closure maps.
    Ri := CoordinateRing(Ai);
    Ap := AffineSpace(Ri);
    // reconstruct the new curve/scheme in Ai
    if iscurve then
	Cp := Curve(Ap,Polynomial(D));
    elif Type(C) eq Prj then
        Cp := Ap;
    else
	Cp := Scheme(Ap,DefiningIdeal(D));
    end if;
    // fix up the domain of the chart
    id := map< Ap -> Ai | [Ap.i: i in [1..Length(Ap)] ]>;
    chart := id * prechart;
    MakeProjectiveClosureMap(Ap,Ambient(C),DefiningPolynomials(chart));
    return Cp,chart;
end intrinsic;

