// *********************************************************************
// * Package: adjoints.mag                                             *
// * =====================                                             *
// *                                                                   *
// * Author: Tobias Beck                                               *
// *                                                                   *
// * Email : Tobias.Beck@oeaw.ac.at                                    *
// *                                                                   *
// * Date  : 06.12.2005                                                *
// *                                                                   *
// *                                                                   *
// * Purpose:                                                          *
// * --------                                                          *
// *                                                                   *
// *   Compute the vector space of (m,n)-adjoints for a projective     *
// *   surface in P^3.                                                 *
// *                                                                   *
// *********************************************************************
freeze;
declare verbose Classify, 1;

import "../../Ring/RngPowAlg/power_series.m" : EvlMod;




// ======================================
// = Auxillary functions (not exported) =
// ======================================

// compute the Ansatz-space for (m,n)-adjoints
// -------------------------------------------
// input:  a polynomial 's' in four variables and integers 'm' and 'n'
// output: a basis for the vector space of forms modulo 's' of degree
//         'n + m * (TotalDegree(s) - 4)'
function Ansatz(s, m, n)
	N := n + m * (TotalDegree(s) - 4);
	S := Parent(s);
	vs := [ S |]; lmon := LeadingMonomial(s);
	if N lt 0 then return vs; end if;
	for zi := 0 to N do
	    for yi := 0 to N-zi do
		for xi := 0 to N-zi-yi do
		    // add monomial if it is part of a basis of O_X(N)
		    mon := (S.4)^zi*(S.3)^yi*(S.2)^xi*(S.1)^(N-zi-yi-xi);
		    if not IsDivisibleBy(mon, lmon) then
		        Append(~vs, mon);
		    end if;
		end for;
	    end for;
	end for;
	return vs;
end function;


// get adjoint conditions for a vector space
// -----------------------------------------
// input:  the basis 'vs' of a vector space of forms, a formal
//         desingularization 'vals' and an integer 'm'
// output: the sequence of conditions for the subspace of those forms that
//         vanish with 'm' times the adjoint order along each formal prime
//         divisor in 'vals'
function GetAdjointConditions(vals, vs, m)
	vprint Classify: "--------- Entering GetAdjointConditions ----------";
	if #vs eq 0 then return vs; end if;
	F := BaseRing(Parent(Representative(vs)));
	IVec := RSpace(Integers(), 4);
	
	// compute adjoint conditions for each valuation
	con := [];
//num := 0;
	for val in vals do
//num := num+1;
		aux, adjord := Explode(val); X, Y, Z, W := Explode(aux);
		Fext := BaseRing(Domain(Z));
//printf "Valuation %o with adjoint order %o and field %o.\n", num, adjord, Fext;
		
		// early abort
		if m*adjord le 0 then
			continue;
		end if;
		
		// approximate series
		approx := []; good := true;
		for s in [X, Y, Z, W] do
			good, app := Expand(s, m*adjord);
//good, app := Expand(s, m*adjord+5);
			if not good then error "series not expandable"; end if;
			Append(~approx, app);
		end for;
//printf "Values = %o\n", approx;
		
		// derive conditions
		evl := EvlMod(approx, m*adjord);
		vecs := [];
		for v in vs do
			// approximate series substitution
			ss := evl(v);
//printf "c_{%o}(%o) + \n", v, ss;
			
			// read of trailing coefficients
			vec := [LeadingCoefficient(Coefficient(ss, 1, i))
			        : i in [0..m*adjord-1]];
			Append(~vecs, vec);
		end for;
		
		// transpose to read off constraints
		V := VectorSpace(Fext, #vs);
		conext := [ V ! v : v in RowSequence(Transpose(Matrix(vecs)))];
		
		// transform relations back to base field
//printf "Constraints before = %o\n", conext;
		contrans := TransformRelations(F, conext);
//printf "Constraints after = %o\n", contrans;
		con := con cat contrans;
	end for;
	
	vprint Classify: "--------- Leaving GetAdjointConditions ----------";
	return con;
end function;


// compute adjoint subspace of a space of forms
// --------------------------------------------
// input:  the basis 'vs' of a vector space of forms and a set of linear
//         constraints 'con'
// output: the subspace fulfilling the linear constraints
function ProduceAdjoints(vs, con)
	if not #con eq 0 then
		
		// there are adjoint conditions
		contrans := Transpose(Matrix(con));
		
		// produce basis of adjoint subspace
		sol := Nullspace(contrans); adjoints := [];
		for b in Basis(sol) do
			adjb := 0;
			for i := 1 to #vs do
				adjb := adjb + b[i]*vs[i];
			end for;
			if not adjb eq 0 then
				Append(~adjoints, adjb);
			end if;
		end for;
	else
		
		// there are no adjoint conditions
		adjoints := vs;
	end if;
	return adjoints;
end function;

// actual function to compute the v.s. of (m,n)-adjoints
function homAdjoints(m,n,p : FormalDesing := 0)
	S := Parent(p); F := BaseRing(S);
	vprint Classify: "------------ Entering HomAdjoints -------------";
	vprintf Classify: "m: %o, n: %o \n", m, n;
	
	// is precomputation available?
	if Type(FormalDesing) eq RngIntElt then
		FormalDesing := ResolveProjectiveSurface(p : AdjComp := true);
	end if;
	
	// produce basis of Ansatz space
	vs := Ansatz(p, m, n);
	vprintf Classify: "dim vs: %o \n", #vs;
	
	// get 'm'-adjoint conditions
	con := GetAdjointConditions(FormalDesing, vs, m);
	
	// produce adjoint basis
	adjoints := ProduceAdjoints(vs, con);
	
	vprint Classify: "------------ Leaving HomAdjoints -------------";
	return adjoints;
end function;

// ======================
// = Exported functions =
// ======================

// compute the vector space of (m,n)-adjoints
intrinsic HomAdjoints(m::RngIntElt, n::RngIntElt, S::Srfc
 : FormalDesing := 0) -> SeqEnum
{
Given a (generally singular) surface S in 'P^3' over a numberfield
and integers m and n. Returns a basis
for the vector space of '(m,n)'-adjoints. A formal
desingularization may also be passed as parameter 'FormalDesing'.
}
	require IsOrdinaryProjective(S) and Dimension(Ambient(S)) eq 3:
	  "S must be a surface in ordinary projective 3-space";
	require (t cmpeq FldRat) or ISA(t,FldAlg) where t is Type(BaseRing(S)):
	  "S must be defined over a number field";

	return homAdjoints(m,n,DefiningPolynomial(S) :
			FormalDesing := FormalDesing);
end intrinsic;
