freeze;
/* Some miscellaneous functions for free group.
 * FreeGroupIsIn(H,x) for H  subgroup of free group should be available as
 * x in H 
 * FreeGroupIndex(G,H) should be used by intrinsic Index(G,H).
 */
import "dfa.m": DFAREC;

intrinsic FreeGroupIndex(G :: GrpFP, H :: GrpFP) -> RngIntElt
{Index of subgroup of free group - natural number or infinity}
  local CT;
  require #Relations(G) eq 0 : "Group must be free";
  require H subset G : "Second argument must be a subgroup of first";
  CT := CosetTable(G, H : Strategy:="RT");
  if IsComplete(CT) then
    return #CT;
  end if;
  return Infinity();
end intrinsic;

intrinsic FreeGroupIsIn(H :: GrpFP, x) -> BoolElt
{Is element x of free group G in subgroup H of G?}
  local G, CT;
  G := Parent(x);
  require #Relations(G) eq 0 : "Group must be free";
  require H subset G : "Second argument must be a subgroup of first";
  CT := CosetTable(G, H : Strategy:="RT");
  return CT(<1,x>) eq 1;
end intrinsic;

intrinsic FreeGenerators(G :: GrpFP, H :: GrpFP) -> SeqEnum
{Free generators of subgroup H of free group G}
  local CT, r, n, T, im;
  require #Relations(G) eq 0 : "Group must be free";
  require H subset G : "Second argument must be a subgroup of first";
  CT := CosetTable(G, H : Strategy:="RT");
  r := Ngens(G);
  n := #CT;
  T := [ Id(G) ]; //elements mapping 1 to cosets
  ndefs := 1;
  for c in [1..n] do
    for g in [1..r] do
      im := CT(<c, G.g>);
      if im gt 0 and not IsDefined(T,im) then
        T[im] := T[c] * G.g;
        ndefs +:=1;
        if ndefs eq n then break c; end if;
      end if;
    end for;
    for g in [1..r] do
      im := CT(<c, G.g^-1>);
      if im gt 0 and not IsDefined(T,im) then
        T[im] := T[c] * G.g^-1;
        ndefs +:=1;
        if ndefs eq n then break c; end if;
      end if;
    end for;
  end for;
  gens := [];
  for c in [1..n] do
    for g in [1..r] do
      im := CT(<c, G.g>);
      if im gt 0 then
         gen := T[c] * G.g * T[im]^-1;
         if gen ne Id(G) then Append(~gens, gen); end if;
      end if;
    end for;
  end for;

  return gens;
end intrinsic;

FreeGeneratorsDFA := function(G, dfa)
//Same as FreeGenerators but subgroup defined by dfa of coset table
  local tab, r, n, T, im;
  assert #Relations(G) eq 0;
  r := Ngens(G);
  tab := dfa`transitions;
  n := #tab;
  T := [ Id(G) ]; //elements mapping 1 to cosets
  ndefs := 1;
  for c in [1..n] do
    for g in [1..r] do
      im := tab[c][g];
      if im gt 0 and not IsDefined(T,im) then
        T[im] := T[c] * G.g;
        ndefs +:=1;
        if ndefs eq n then break c; end if;
      end if;
    end for;
    for g in [1..r] do
      im := tab[c][r+g];
      if im gt 0 and not IsDefined(T,im) then
        T[im] := T[c] * G.g^-1;
        ndefs +:=1;
        if ndefs eq n then break c; end if;
      end if;
    end for;
  end for;
  gens := [];
  for c in [1..n] do
    for g in [1..r] do
      im := tab[c][g];
      if im gt 0 then
         gen := T[c] * G.g * T[im]^-1;
         if gen ne Id(G) then Append(~gens, gen); end if;
      end if;
    end for;
  end for;

  return gens;
end function;

intrinsic CosetTableToDFA(CT :: Map) -> Rec
{Return DFA corresponding to coset table CT. Accepted words are words in
the subgroup used to define CT}
  local G, tab, row, ng;
  G := Group(CT);
  ng := Ngens(G);
  tab := [];
  for s in [1..#CT] do
    row := [];
    for i in [1..ng] do
      Append( ~row, CT(< s, G.i >) );
    end for;
    for i in [1..ng] do
      Append( ~row, CT(< s, G.i^-1 >) );
    end for;
    Append(~tab, row);
  end for;
  return rec< DFAREC | initial := {1}, final := {1},
                       alphabetsize := 2*ng, transitions := tab >;
end intrinsic;
  
intrinsic SubgroupsMeet(G :: GrpFP, H :: GrpFP, K :: GrpFP) -> GrpFP
{H meet K for H, K subgroups of free group G}
  local dfaH , dfaK, dfameet;
  require #Relations(G) eq 0 : "For type GrpFP, group must be free";
  require H subset G : "Second argument must be a subgroup of first";
  require K subset G : "Third argument must be a subgroup of first";
  dfaH := CosetTableToDFA( CosetTable(G, H : Strategy:="RT") );
  dfaK := CosetTableToDFA( CosetTable(G, K : Strategy:="RT") );
  dfameet := MinimizeDFA( MeetDFA(dfaH, dfaK) );
  return sub< G | FreeGeneratorsDFA(G, dfameet) >; 
end intrinsic;

intrinsic SubgroupElementsCT(CT :: Map, len :: RngIntElt) -> SetEnum
{Given a (not necessarily complete) coset table for a subgroup H of a
 finitely presented group G, return the elements of G of length at most
 len that are in H}
  local dfa, G;
  dfa := CosetTableToDFA(CT);
  G := Group(CT);
  dfa`alphabet := [G.i : i in [1..Ngens(G)]] cat [G.i^-1 : i in [1..Ngens(G)]];
  return LanguageDFA(dfa, 0, len : words:=true );
end intrinsic;

intrinsic SubgroupElements(G :: GrpFP, H :: GrpFP, len :: RngIntElt) -> SetEnum
{Given a subgroup H of a finitely presented group G, form the coset
 table for H in G, and use it to return the elements of G of length at most
 len that are in H. If options to CosetTable need to be used, then call
 CosetTable first followed by SubgroupElementsCT.}
  local CT;
   if #Relations(G) eq 0 then //G is free so incompelte coset table OK
     CT := CosetTable(G, H : Strategy:="RT");
   else
     CT := CosetTable(G, H);
     if not IsComplete(CT) then
       "WARNING: Coset table is incomplete!";
     end if;
   end if;
   return SubgroupElementsCT(CT, len);
end intrinsic;
