/* 
  This file contains the functions that compute generators
  for the subgroups of the classical group G that are used in the 
  rewriting process.
  
  PSubgroup returns the p-core of the stabilizer of 
  <e_2,...,e_m,f_m,...,f_1>.
  
*/
  
import "bbclassical.m":B, BBType, BBField, BBStandardGenerators, 
  BBDimension;
import "elements.m":X_, Tf1aw;    

/*
  PSubgroup is implemented for the various types of 
  classical groups. The parameter Transpose is used in the SL case 
  when we need also
  compute the transpose of this subgroup.
*/

  
PSubgroup := function( G : Transpose := false )
    
    if not Transpose and assigned G`PSubgroup then
        return G`PSubgroup[1],  G`PSubgroup[2];
    elif Transpose and assigned G`PSubgroupTranspose then
        return G`PSubgroupTranspose[1],  G`PSubgroupTranspose[2];
    end if;  
    
    type := BBType( G );
    gens := BBStandardGenerators( G );
    dim := BBDimension( G );
    q := #BBField( G );
      
    _, p, d := IsPrimePower( q );  
    S := gens[1]; V := gens[2]; D := gens[4]; 
    U := gens[5]; 
    
    X1 := gens[6]; X2 := gens[7];
    T := gens[3];
    
    V1 := V*U; pV1 := B.2*B.5;
    
    if type eq "Omega+" and dim eq 4 then 
        D1 := gens[4]*gens[10]; pD1 := B.4*B.10;
    else
        D1 := D^V1; pD1 := B.4^pV1;
    end if;
        
    
    if not Transpose then
        T1 := X1; pT1 := B.6;
    else
        T1 := X1^U; pT1 := B.6^B.5;
    end if;
    
    pgens := []; bbgens := [];
    
    if type eq "SL" then 
        up := dim-1; 
    elif type eq "Omega-" then 
        up := (dim-4) div 2;
    else
        up := (dim-2) div 2;
    end if;
    
    exp := case< type | "SL": -1, "Omega+": -1, 
           "Omega-": -1, "Omega": -1, default: 1 >;
    
    for i in [1..up] do
        T2 := T1; pT2 := pT1;
        for j in [1..d] do
            Append( ~pgens, T2 );
            Append( ~bbgens, pT2 );
            T2 := (T2^(D1)); pT2 := (pT2^pD1);
        end for;
        D1 := D1^V1; pD1 := pD1^pV1;
        T1 := (T1^V1)^exp; pT1 := (pT1^pV1)^exp;
    end for;
    
    // if SL then we are done.
    
    if type eq "SL" and not Transpose then
        P := sub< Parent( D ) | pgens >;
        G`PSubgroup := < P, bbgens >;
        return P, bbgens;
    elif  type eq "SL" and  Transpose then
        P := sub< Parent( D ) | pgens >;
        G`PSubgroupTranspose := < P, bbgens >;
        return P, bbgens;
    end if;
    
    T1 := X2; pT1 := B.7;
    if type in { "Omega+", "Omega-" } then 
        T1 := T1^-1; pT1 := pT1^-1;
    end if;
    
    D1 := D^V1; pD1 := B.4^pV1;
    
    for i in [1..up] do
        T2 := T1; pT2 := pT1;
        for j in [1..d] do
            Append( ~pgens, T2 );
            Append( ~bbgens, pT2 );
            T2 := (T2^(D1)); pT2 := (pT2^pD1);
        end for;
        D1 := D1^V1; pD1 := pD1^pV1;
        T1 := (T1^V1)^exp; pT1 := (pT1^pV1)^exp;
    end for;
    
    T1 := T; pT1 := B.3;
    
    // add the transvection subgroup. orthogonal groups don't have one.
    
    if not type in { "Omega", "Omega+", "Omega-" } then
        
        if type eq "SU" then 
            d0 := d div 2;
        else
            d0 := d;
        end if;
        
        for i in [1..d0] do
            Append( ~bbgens, pT1 );
            Append( ~pgens, T1 );
            T1 := T1^D; pT1 := pT1^B.4;
        end for; 
        
    end if;
    
    // add T_{f1,w} for SU( odd, q ) and Omega( odd, q )
    
    if IsOdd( dim ) then
        
        X := gens[8]; pX := B.8; 
        
        if type eq "SU" then 
            D1 := D^V; pD1 := B.4^B.2;
        elif type eq "Omega" then
            D1 := D; pD1 := B.4;
        end if;
        
        for i in [0..d-1] do
            Append( ~bbgens, pX );
            Append( ~pgens, X );
            X := (X^D1); pX := pX^pD1;
        end for;
        
    end if;  
    
    // T_{f1,w2} in Omega-( n, q )
      
    if type eq "Omega-" then  
        X := gens[8]; pX := B.8; 
        D1 := D; pD1 := B.4;
        
        for i in [0..2*d-1] do
            Append( ~bbgens, pX );
            Append( ~pgens, X );
            X := (X^D1); pX := pX^pD1;
        end for;
        
    end if;  
      
    P := sub< Parent( D ) | pgens >;
    G`PSubgroup := < P, bbgens >;
    return P, bbgens;
end function;
    
/* 
  The following constructs stabilizer of E1 in G. Transpose is needed
  in the SL case. When "on" then the transpose of the stabilizer
  is returned.
*/
    
StabilizerOfE1 := function( G : Transpose := false )
      
      if not Transpose and assigned G`StabilizerOfE1 then
          return G`StabilizerOfE1;
      elif Transpose and assigned G`StabilizerOfE1Transpose then
          return G`StabilizerOfE1Transpose;
      end if; 
            
     gens := BBStandardGenerators( G );
      type := BBType( G );
      dim := BBDimension( G );
      q :=    #BBField( G );
        
        // low dim groups have no gens[5].
        // the following boolean will show which ones do.

      hasgens5 := not < type, dim > in { <"SU",4>, 
                          <"Sp",4>, 
                          <"SU",5>, 
                          <"Omega-",4>,
                          <"Omega-",6>,
                          <"Omega",5>,
                          <"Omega",3> };

      if type eq "SL" and dim eq 2 and not Transpose then
          
          stabgens := [ gens[3]^gens[1], gens[4] ];
          
      elif type eq "SL" and dim eq 2 and Transpose then
          
          stabgens := [ gens[3], gens[4] ];
          
      elif type eq "SU" and dim eq 3 then
          
          stabgens := [ gens[4],gens[10],gens[3]^gens[1], gens[8]^gens[1]]; 
          
      elif type eq "Omega+" and dim eq 4 then
                    
          stabgens := [ gens[4], gens[6]^gens[5], gens[7]^gens[1], gens[10]];
          
      elif type eq "Omega-" and dim eq 4 and IsOdd( q ) then
          
          stabgens := [ gens[4], gens[11], gens[8]^gens[1]];

      elif  type eq "Omega-" and dim eq 4 and IsEven( q ) then
          
          stabgens := [ gens[4], gens[11], 
                        gens[8]^gens[2]*gens[1]^(gens[2]^-1)];
          
      elif type eq "Omega" and dim eq 3 then
          
          stabgens := [ gens[4], gens[8]^gens[1]];
          
      elif type eq "Omega-" and IsEven( q ) then
          
          stabgens := [ 
                        gens[1]^gens[2]^-1,
                        gens[2]*gens[5],
                        gens[3]^gens[2],
                        gens[4],
                        gens[4]^gens[2],
                        case< hasgens5 | true:  gens[5]^gens[2],
                        default: One( G )>,
                        gens[6]^gens[2],
                        //gens[7]^gens[2],
                        //gens[8]^gens[2],
                        //gens[9]^gens[2],
                        gens[10],
                        gens[3]^gens[5],
                        gens[6]^gens[2]^-1];
          
    elif not Transpose then
        

        stabgens := [ 
	          // the expression for gens[1] is different in Omega+
                      case< type | "Omega+": gens[1]^gens[2], 
                      default: gens[1]^gens[2]^-1 >,
                      gens[2]*gens[5],
                      gens[3]^gens[2],
                      gens[4],
                      gens[4]^gens[2],
		  // low dim groups have no gens[5] 
                      case< hasgens5 | true: gens[5]^gens[2],
                      default: One( G )>,
                      gens[6]^gens[2],
		      case< <type,dim> | <"Omega+",4>: gens[7]^gens[2],
                                         default: One( G )>,
                      case< <type,dim> | <"Omega+",6>: gens[7]^gens[2],
                                         default: One( G )>,	  
                      gens[8]^gens[2],
                      gens[9]^gens[2],
                      gens[10],
                      gens[3]^gens[5],
                      gens[7]^gens[1],
                      gens[8]^gens[1],
                      gens[9]^gens[1] ];

    elif Transpose then
    
        stabgens := [ gens[3]^gens[2]^2,
                      gens[2]*gens[5],
                      gens[1]*gens[2],
                      gens[4]^gens[2],
                      gens[3] ];
        
    end if;
    
    stab := sub< GL( Degree( G ), BaseRing( G )) | stabgens >;
    
    if not Transpose then
        G`StabilizerOfE1 := stab;
    else
        G`StabilizerOfE1Transpose := stab;
    end if;
    
    return stab;
end function;
    
    
/* 
  the following is a function that returns part of the generating
  set of PSubgroup, namely elements of the form T_{f1,b}. needed
  in SmallerMatrix.
*/
  
RootElements := function( G : Transpose := false ) 
    
    dim := BBDimension( G );
    q := #BBField( G ); 
    _, p, d := IsPrimePower( q );
    type := BBType( G );
    gens := BBStandardGenerators( G );
    T := gens[3]; V := gens[2]; U := gens[5];
    
    P := PSubgroup( G );
    
    if type eq "SL" then
        return [ P.x : x in [1..(dim-2)*d+1 by d ]];
    end if;
    
    
    if type in {"SL","Sp","Omega+"} or ( type eq "SU" and IsEven( dim )) then
        
        r1 := ((dim-2) div 2)*d+1; r2 := (dim-3)*d+1;
        r3 := ((dim-3) div 2)*d+1; r4 := 1;
        
    elif type eq "Omega" or ( type eq "SU" and IsOdd( dim )) then
        
        r1 := ((dim-2) div 2)*d+1; r2 := (dim-4)*d+1;
        r3 := ((dim-4) div 2)*d+1; r4 := 1;
        
    elif type eq "Omega-" then
        
        r1 := ((dim-4) div 2)*d+1; r2 := (dim-5)*d+1;
        r3 := ((dim-6) div 2)*d+1; r4 := 1;
        
    end if;
    
    range2 := [r2..r1 by -d];
    range1 := [r4..r3 by d];
    
    if type in { "Omega+", "SL", "Sp", "Omega-" } 
         or ( type eq "SU" and IsEven( dim )) then
        roots := [];
    elif type eq "SU" and IsOdd( dim ) then
        roots := [ P.((dim-3)*d + d div 2 +1)];
    elif type eq "Omega" then
        roots := [ P.((dim-3)*d+1)];
    end if;
    
    roots := roots cat [ P.x : x in range1 ];
    
    if type eq "Omega-" then
        roots := roots cat [ gens[8], gens[9] ];
    end if;
    roots := roots cat [ P.x : x in range2 ];
    
    return roots;
end function;
    
     
DiagonalElement := function( type, dim, q )
    
    if type eq "SL" then
        
        p := B.2^-1*B.1;
        pr := B.4^-1;
        
        for i in [1..dim-2] do
            pr := pr*((B.4^-1)^p);
            p := p*B.2^-1*B.1;
        end for;
        
        return pr;
        
    elif type eq "Sp" then
        
        p := B.2*B.5;
        pd := B.4^B.5; pr := pd;
        
        for i in [1..(dim-4)/2] do
            pd := pd^(B.2*B.5);
            pr := pr*pd;
        end for;
        
        pr := pr^Round((q-2)/2);
        return pr;
        
    elif type eq "SU" and dim eq 3 then
        
        return B.4;
        
    elif type eq "SU" and IsEven( dim ) then
        
        progY1 := One( B ); progY2 := B.4*B.4^(B.1*B.1^B.2);
        
        for i in [1..(dim-2)/2] do
            progY1 := progY1*progY2; progY2 := progY2^(B.2*B.5);
        end for;
        
        return progY1;
        
    elif type eq "SU" and IsOdd( dim ) then
        
        progZ1 := (B.4*B.4^B.1)^-1;
        progZ := One( B );
        
        for i in [1..(dim-1)/2] do
            progZ := progZ*progZ1;
            progZ1 := progZ1^B.2;
        end for;
        
        progZ := progZ*B.4^((2*dim div 2));
        
        
        return progZ;
        
    elif type eq "Omega+" then
        
        progZ := One( B );
        progZ1 := B.4^((q-1) div 2);
        V := B.2*B.5;
        
        for i in [1..(dim-2)/2] do
            progZ := progZ*progZ1;
            progZ1 := progZ1^V;
        end for;
        
        return progZ;
        
    elif type eq "Omega-" then
        
        q1 := (q-1) div 2; dim1 := dim div 2-2;
        pd1 := B.11^q1;
        pd2 := (B.11^q1)^B.2^-1;
        for i in [1..dim1-1] do
            pd1 := pd1*pd2;
            pd2 := pd2^B.2^-1;
        end for;
        
        pd1 := pd1*(B.11^B.2)^Round((q+1)/2-q1*dim1);
        return pd1;
        
    elif type eq "Omega" then
        
        return One( B );
        
    end if;
end function;
    
        
