freeze;

declare attributes FldFunAb:
  D,  // the divisor used to define the field,
  U,  // the subgroup 
  C,  // the conductor - if known
  F,  // the function field - if known
  g,  // the genus - if known
  N,  // the number of places of degree 1
  dk, // the degree of the ECF
  disc,// the discrimant divisor
  witts; // a list of witt vectors, entries set for those components 
  	 // which are ASW extensions
	 // primitive element of Kummer extension otherwise 

////////////////////////////////////////////////////////////////////
//                                                                // 
//                       Creation functions                       // 
//                                                                //
////////////////////////////////////////////////////////////////////

/* cannot get the divisir from the map...
intrinsic AbelianExtension(m::map[GrpAb, DivFun]) -> FldFunAb
  {}
end intrinsic;
*/

intrinsic AbelianExtension(D::DivFunElt, U::GrpAb) -> FldFunAb
  {Creates the abelian extension defined by D and U.}
  r, mr := RayClassGroup(D);
  require U subset r:
    "The 2nd argument must be a subgrop of the ray class group modulo the 1st";
  q, mq := quo<r|U>;
  require IsFinite(q) :
    "The subgroup must be of finite index in the ray class group";

  A := HackobjCreateRaw(FldFunAb);

  A`D := D;
  A`U := U;
  return A;
end intrinsic;

intrinsic HackobjPrintFldFunAb(A::FldFunAb, l::MonStgElt) 
  {}
  r, mr := RayClassGroup(A`D);
  q, mq := quo<r|A`U>;

  "Abelian extension of type", AbelianInvariants(q);
  any := false;
  if assigned A`g then
    printf "of genus %o, ", A`g;
    any := true;
  end if;
  if assigned A`N then
    printf "with %o rational places, ", A`N;
    any := true;
  end if;
  if assigned A`dk then
    printf "and exact constant field of degree %o", A`dk;
    any := true;
  end if;
  if any then
    printf "\n";
  end if;
  "Defined modulo", A`D;
  "over", FunctionField(A`D);
  if l eq "Maximal" then
    if assigned A`C then
      "Of conductor", A`C;
    end if;
    if assigned A`disc then
      "And discrimant divisor", A`disc;
    end if;
    if assigned A`F then
      "As a function field", A`F;
    end if;
  end if;
end intrinsic;   

intrinsic FunctionField(A::FldFunAb:WithAut := false) -> FldFun
  {Computes an explicit representation of A as a finite extension.}
  if assigned A`F then
    if (WithAut and Type(A`F) eq SeqEnum) or 
    					(not WithAut and Type(A`F) eq FldFun) then
	return A`F, A`witts;
    end if;
    if not WithAut and Type(A`F) ne FldFun then
	return FunctionField([DefiningPolynomial(x) : x in A`F] : Check := false), 								A`witts;
    end if;
    if WithAut and Type(A`F) ne SeqEnum then
	return [FunctionField(x) : x in DefiningPolynomials(A`F)];
    end if;
  end if;
  F, _, witts := InternalRayClassField(A`D, A`U:WithAut := WithAut);
  A`F := F;
  A`witts := witts;
  return F;
end intrinsic;    

intrinsic Conductor(A::FldFunAb) -> DivFunElt
  {The conductor of A}
  if assigned A`C then
    return A`C;
  end if;
  A`C := Conductor(A`D, A`U);
  return A`C;
end intrinsic;

intrinsic DiscriminantDivisor(A::FldFunAb) -> DivFunElt
  {The discrimant divisor of A}
  if assigned A`disc then
    return A`disc;
  end if;
  A`disc := DiscriminantDivisor(A`D, A`U);
  return A`disc;
end intrinsic;

intrinsic DegreeOfExactConstantField(A::FldFunAb) -> RngIntElt
  {The degree of the exact constant field of A}
  if assigned A`dk then
    return A`dk;
  end if;
  A`dk := DegreeOfExactConstantField(A`D, A`U);
  return A`dk;
end intrinsic;

intrinsic Genus(A::FldFunAb) -> RngIntElt
  {The genus of A}
  if assigned A`g then
    return A`g;
  end if;
  A`g := Genus(A`D, A`U);
  return A`g;
end intrinsic;

intrinsic MaximalAbelianSubfield(F::FldFun: 
             Cond := false, AS := false) -> FldFunAb
  {Computes the maximal abelian subfield of simple extension F as an abelian extension}
   
  D, U := NormGroup(F:Cond := Cond, AS := AS);  
  return AbelianExtension(D, U);
end intrinsic;

intrinsic DecompositionType(A::FldFunAb, p::PlcFunElt) -> SeqEnum
  {The decomposition type of p in A as a list of pairs <f, e>}
  return DecompositionType(A`D, A`U, p);  
end intrinsic;

intrinsic NumberOfPlacesOfDegreeOne(A::FldFunAb) -> RngIntElt
  {The number of places of degree one in A}
  if assigned A`N then
    return A`N;
  end if;
  A`N := NumberOfPlacesOfDegreeOne(A`D, A`U);
  return A`N;
end intrinsic;

intrinsic Degree(A::FldFunAb) -> RngIntElt
  {The relative degree of A}
  r, mr := RayClassGroup(A`D);
  q, mq := quo<r|A`U>;
  return #q;
end intrinsic;

intrinsic BaseField(A::FldFunAb) -> FldFunG
  {The coefficient field of A}
  return FunctionField(A`D);
end intrinsic;


function make_compatible(A, B)
  D := LCM(A`D, B`D);
  RD, mRD := RayClassGroup(D);
  RA, mRA := RayClassGroup(A`D);
  RB, mRB := RayClassGroup(B`D);
  //hA := hom<RD -> RA | [RD.i @ mRD @@ mRA : i in [1..Ngens(RD)]]>;
  //hB := hom<RD -> RB | [RD.i @ mRD @@ mRB : i in [1..Ngens(RD)]]>;
  hA := InducedMap(mRD, mRA, IdentityHomomorphism(Codomain(mRD)), D);
  hB := InducedMap(mRD, mRB, IdentityHomomorphism(Codomain(mRD)), D);
  return D, (A`U) @@ hA, (B`U) @@ hB;
end function;

intrinsic 'subset'(A::FldFunAb, B::FldFunAb) -> BoolElt
  {Tests if A is a subfield of B, for abelian extensions of the same base field}

  require BaseField(A) cmpeq BaseField(B):
    "Both fields must be defined over the same base";

  D, UA, UB := make_compatible(A, B);
  return UB subset UA;
end intrinsic

intrinsic 'eq'(A::FldFunAb, B::FldFunAb) -> BoolElt
  {Tests if A equals B, for abelian extensions of the same base field}

  require BaseField(A) cmpeq BaseField(B):
    "Both fields must be defined over the same base";

  D, UA, UB := make_compatible(A, B);
  return UB eq UA;
end intrinsic

intrinsic 'meet'(A::FldFunAb, B::FldFunAb) -> FldFunAb
  {The intersection of A and B, for abelian extensions of the same base field}

  require BaseField(A) cmpeq BaseField(B):
    "Both fields must be defined over the same base";

  D, UA, UB := make_compatible(A, B);
  return AbelianExtension(D, sub<RayClassGroup(D)| UB,  UA>);
end intrinsic

intrinsic '*'(A::FldFunAb, B::FldFunAb) -> FldFunAb
  {The compositum of A and B, for abelian extensions of the same base field}

  require BaseField(A) cmpeq BaseField(B):
    "Both fields must be defined over the same base";

  D, UA, UB := make_compatible(A, B);
  return AbelianExtension(D, UB meet UA);
end intrinsic

function asw_max_ord(F, u, finite)

// need the Witt elt u, either stored or from the Witt tower 
// probably best to store u since it takes less space
// attribute of Witt elts, one for each component

    if finite then
	return MaximalOrderFinite(F);
    else
	return MaximalOrderInfinite(F);
    end if;
end function;

function maximal_order(A, finite, Al, Partial)
    F := FunctionField(A);

    //check attribute

    if finite and assigned F`MaximalOrderFinite then
	return F`MaximalOrderFinite;
    end if;

    if not finite and assigned F`MaximalOrderInfinite then
	return F`MaximalOrderInfinite;
    end if;

    DA := DiscriminantDivisor(A);
    DA := DivisorGroup(RationalExtensionRepresentation(FunctionField(DA)))!DA;
    if finite then 
	max_ord := MaximalOrderFinite;
	eq_ord := EquationOrderFinite;
	eq_ord_F := EquationOrderFinite(F);
	DA := FiniteDivisor(DA);
	I := Ideals(DA);
    else
	max_ord := MaximalOrderInfinite;
	eq_ord := EquationOrderInfinite;
	eq_ord_F := EquationOrderInfinite(F);
	DA := InfiniteDivisor(DA);
	_, I := Ideals(DA);
    end if;

    C := [FunctionField(f) : f in DefiningPolynomials(F)];
    for i in [1 .. #C] do
	Embed(C[i], F, F.i);
    end for;

    p := Characteristic(F);
    // components are all cyclic extensions
    d := [];
    for i in [1 .. #C] do
	c := C[i];
	deg := Degree(c);
	if deg mod p ne 0 then
if not IsExport() then
"ne Kummer";
end if;
	    // Kummer 
	    // check degrees or coeff rings? 
	    // These might be isomorphic but I'm not sure what the equality will do
	    PAw := Parent(A`witts[i]);
	    if true and PAw ne c then 
		//assert AbsoluteDegree(c) ne AbsoluteDegree(PAw);
		//assert AbsoluteDegree(c)*Degree(CoefficientField(PAw)) eq AbsoluteDegree(PAw);
		assert CoefficientField(c) ne CoefficientField(PAw);
		assert CoefficientField(c) eq CoefficientField(CoefficientField(PAw));
		ac := RationalExtensionRepresentation(CoefficientField(PAw));
		kum_abs := ext<ac | DefiningPolynomial(PAw)>;
		Embed(kum_abs, PAw, PAw.1);

		// Using primes and internal as for number fields
		// Is this faster than the meet below?
		/*
		if false then
		primes := Factorization(Discriminant(eq_ord(c))*max_ord(ac));
		primes;
		m := InternalMaximalOrderUsingKummer(eq_ord(c), eq_ord(PAw), eq_ord(kum_abs), primes);
		cc := ext<CoefficientField(c) | DefiningPolynomial(c)>;
		assert Discriminant(m) eq Discriminant(max_ord(cc));
if not IsExport() then
"ne Disc check passed on Internal";
end if;
		end if;
		*/

		// meet
		kum_abs_m := max_ord(kum_abs);
		// BUT NEED TO COERCE OVER COEFF RING PAw first!!
		pb := PseudoBasis(Module(kum_abs_m));
		M := Module([<CoefficientRing(eq_ord(PAw))!!t[1], RM!Eltseq(t[2])> : t in pb]) where RM := RModule(CoefficientRing(eq_ord(PAw)), Degree(PAw));
		PAwm := Order(eq_ord(PAw), M);
		Embed(c, PAw, A`witts[i]);
		m := PAwm meet c;
if not IsExport() then 
		Factorization(Discriminant(m));
		assert Discriminant(m) eq Discriminant(max_ord(c));
"ne Disc check passed on meet : need to remove this!";
end if;
	    else 
		// c is kummer so use it directly
		m := max_ord(c);
	    end if;
	    Append(~d, m);
	elif deg eq p then
if not IsExport() then
"ne Artin--Schreier";
end if;
	    // AS
	    Append(~d, max_ord(c));
	else 
if not IsExport() then
"ne Artin--Schreier-Witt";
end if;
	    // ASW?
	    assert #Factorization(deg) eq 1;
	    assert IsDefined(A`witts, i);
	    Append(~d, asw_max_ord(c, A`witts[i], finite));
	end if;
    end for;

    g := [F|1 ];
    abs := IsAbsoluteOrder(eq_ord_F);
    B := BaseRing(eq_ord_F);
    if not abs then
	B := MaximalOrder(B);
	eq_ord_F := ext<B | DefiningPolynomials(eq_ord_F)>;
	id := [ideal<B | 1> ];
    end if;
    for i in [1..#C] do
	o := d[i];
	g := [ i*F!o.j : i in g, j in [1..Degree(o)]];
	if not abs then
	    pm := PseudoBasis(Module(o));
	    id := [ i*pm[j][1] : i in id, j in [1..Degree(o)]];
	end if;
    end for;

    M := KSpace(FieldOfFractions(B), Degree(F));
    if not abs then
	m := [ <id[i], M!Eltseq(g[i])> : i in [1..#g] ];
	vprint MaximalOrder, 2: "... done, the normal form";
	m := Module(m);
	vprint MaximalOrder, 2: "... done, the new order";
	g := Order(eq_ord_F, m);
    else
	m := Matrix([ M!Eltseq(FieldOfFractions(eq_ord_F)!g[i]) : i in [1..#g] ]);
	M := MatrixRing(B, Degree(F));
	if finite then
	    den := Denominator(m);
	else
	    min_val := Minimum([Degree(Denominator(c)) - Degree(Numerator(c)) : c in Eltseq(m)]);
	    den := FieldOfFractions(B).1^(min_val);
	end if;
	m *:= den;
	g := Order(eq_ord_F, M!m, B!den);
    end if;
    U := Parent(Discriminant(d[1]));
    d := &+ [ U | Discriminant(i) : i in d];
    if #C eq 1 or (#C eq 2 and (d cmpeq 1 or not abs and d eq BaseRing(g)!1*BaseRing(g))) or (abs and B!Discriminant(g) eq B!Minimum(I) or (not abs and Discriminant(g) eq I)) then
	SetOrderMaximal(g, true);
    else
	if Partial then
	    "Failed to get the maximal order this way!\n";
	    "Disc is ", Discriminant(g);
	    "It ought to ", DA;
	    "So sqrt(index) is ", Sqrt(Discriminant(g)/I);
	    return g;
	else
	    //Discriminant(eq_ord_F);
	    //Discriminant(g);
	    g := MaximalOrder(g:Discriminant := abs select B!Minimum(I) else I);
	end if;
    end if;

    if finite then 
	F`MaximalOrderFinite := g;
    else
	F`MaximalOrderInfinite := g;
    end if;
    return g;

end function;

intrinsic MaximalOrderFinite(A::FldFunAb : Al := "Default", Partial := false) -> RngFunOrd
{The finite maximal order of A};
    return maximal_order(A, true, Al, Partial);
end intrinsic;

intrinsic MaximalOrderInfinite(A::FldFunAb : Al := "Default", Partial := false) -> RngFunOrd
{The infinite maximal order of A};
    return maximal_order(A, false, Al, Partial);
end intrinsic;
